from prompts import ask
from io import StringIO
from tempfile import TemporaryFile


def test_ask():
    output = StringIO()

    with TemporaryFile("w+") as file:
        file.write("hello world\r")
        file.flush()
        file.seek(0)
        assert ask("test", output_file=output, input_file=file) == "hello world"

    assert (
        output.getvalue()
        == "\r?\x1b[0m \x1b[1mtest \x1b[1m\x1b[30m›\x1b[0m \x1b[0m\x1b[37m\x1b[0m\x1b[32m\x1b[2m\x1b[4mh\x1b[0m\x1b[32m\x1b[2m\x1b[4me\x1b[0m\x1b[32m\x1b[2m\x1b[4ml\x1b[0m\x1b[32m\x1b[2m\x1b[4ml\x1b[0m\x1b[32m\x1b[2m\x1b[4mo\x1b[0m\x1b[32m\x1b[2m\x1b[4m \x1b[0m\x1b[32m\x1b[2m\x1b[4mw\x1b[0m\x1b[32m\x1b[2m\x1b[4mo\x1b[0m\x1b[32m\x1b[2m\x1b[4mr\x1b[0m\x1b[32m\x1b[2m\x1b[4ml\x1b[0m\x1b[32m\x1b[2m\x1b[4md\x1b[0m\n\x1b[F\x1b[1m\x1b[32m✔\x1b[0m\x1b[0m \x1b[1mtest \x1b[1m\x1b[30m›\x1b[0m \x1b[0m\x1b[32m\x1b[2m\x1b[4mhello world\x1b[0m\n"
    )



from __future__ import annotations

from typing import IO, Callable, Union

from colorama import Fore, Style
from typing_extensions import TypeAlias

from ._styles import UNDERLINE, ITALIC
from .input import take_input
from .prompt import Prompt

ValidatorResult: TypeAlias = Union[str, bool]
Validator: TypeAlias = Callable[[str], ValidatorResult]

__all__ = ("ask", "ValidatorResult", "Validator")


def ask(
    question_str: str,
    *,
    default: str | None = None,
    initial: str | None = None,
    input_file: IO[str] | None = None,
    output_file: IO[str] | None = None,
    validate: Validator | list[Validator] | None = None,
    default_invalid: str = "Invalid response.",
) -> str | None:
    """
    Ask a question with a prompt.

    Args:
        question_str: Prompt to show the user.
        initial: Initial value, as shown in `take_input`.
        input_file: File descriptor to read input from. `sys.stdin` by default.
        output_file: File descriptor to write output to. `sys.stdout` by default.
        validate: Validator(s) to be used for validation of the input.
        default_invalid: Default message to show the user if a validator returns `False`.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C

    Returns:
        The inputted value.
    """
    validators = (
        validate if isinstance(validate, list) else [validate] if validate else []
    )

    with Prompt(question_str, output_file=output_file) as prompt:
        while True:
            if default and (not initial):
                initial = default

            text = take_input(
                input_file,
                output=output_file,
                initial=initial,
                color_sequence=Fore.GREEN + Style.DIM + UNDERLINE,
                initial_color=(
                    Fore.BLACK + Style.BRIGHT + ITALIC
                    if not prompt.answer
                    else Fore.RED + Style.BRIGHT + UNDERLINE
                ),
            )
            if (text == "") and default:
                text = default
            error: str | bool | None = None
            for validator in validators:
                failed_message = validator(text)

                if failed_message is not True:
                    if failed_message is False:
                        error = default_invalid
                    else:
                        error = failed_message

            prompt.answer = text
            if error:
                prompt.error(error)
                initial = text
            else:
                prompt.done()
                return text



from __future__ import annotations
from typing import Any

import click
from click.types import convert_type

from .question import ValidatorResult, ask
from .radio import ask_confirm, ask_radio

__all__ = ("PrettyOption",)


def prompt(
    text: str,
    default: Any | None = None,
    type: click.ParamType | Any | None = None,
) -> Any:
    if isinstance(type, click.Choice):
        index = ask_radio(text, type.choices)
        return type.choices[index]

    if type is click.BOOL:
        assert isinstance(default, bool)
        return ask_confirm(text, default=default)

    value_proc = convert_type(type, default)
    result: Any = None

    def validate(value: Any) -> ValidatorResult:
        try:
            nonlocal result
            result = value_proc(value)
        except click.UsageError as e:
            return f"{e.message}"

        return True

    ask(text, validate=validate, default=str(default) if default else None)
    return result


class PrettyOption(click.Option):
    """Option class for Click integration."""

    def prompt_for_value(self, ctx: click.Context) -> Any:
        assert self.prompt is not None
        return prompt(self.prompt, self.default, self.type)



from __future__ import annotations
from typing import Final


def right(amount: int) -> str:
    return f"\033[{amount}C"


def left(amount: int) -> str:
    return f"\033[{amount}D"


def up(amount: int) -> str:
    return f"\033[{amount}A"


def down(amount: int) -> str:
    return f"\033[{amount}B"


UNDERLINE: Final[str] = "\033[4m"
PREVIOUS_FIRST: Final[str] = "\033[F"
BOLD: Final[str] = "\033[1m"
ITALIC: Final[str] = "\033[3m"
CLEAR: Final[str] = "\033[K"

CTRL_C: Final[str] = "\x03"
ARROW: Final[str] = "\x1b"
BACKSPACE: Final[str] = "\x7f"



from __future__ import annotations
import sys
from typing import IO, Literal, Sequence

from colorama import Fore, Style

from ._getch import getch
from ._styles import CTRL_C, PREVIOUS_FIRST, UNDERLINE, left
from .prompt import Prompt

__all__ = ("radio", "ask_radio", "confirm", "ask_confirm")


def radio(
    items: Sequence[str],
    *,
    multi_line: bool = False,
    input_file: IO[str] | None = None,
    output_file: IO[str] | None = None,
    selected_color: str = Fore.GREEN + Style.DIM + UNDERLINE,
    transient: bool = False,
) -> int:
    """
    Display a radio selection.

    Args:
        items: Sequence of available selections. Must have at least 2 values.
        input_file: File descriptor to read input from. `sys.stdin` by default.
        output_file: File descriptor to write output to. `sys.stdout` by default.
        selected_color: Color to use for selected values. Green by default.
        transient: Whether to delete the radio selection after it's done.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C

    Returns:
        Index of the selected item.
    """
    input_file = input_file or sys.stdin
    output_file = output_file or sys.stdout
    selected = 0

    if len(items) < 2:
        raise ValueError("items must have a length of at least 2")

    def render_questions():
        if multi_line:
            for index, question in enumerate(items):
                if selected == index:
                    print(
                        f" - {selected_color}{question}{Style.RESET_ALL}",
                        flush=True,
                        file=output_file,
                    )
                else:
                    print(f" - {question}", flush=True, file=output_file)
        else:
            for index, question in enumerate(items):
                color = selected_color if index == selected else ""
                print(
                    f"{color}{question}{Style.RESET_ALL}",
                    flush=True,
                    end="",
                    file=output_file,
                )

                if index != len(items) - 1:
                    print(
                        f"{Style.BRIGHT}{Fore.BLACK} / {Style.RESET_ALL}",
                        flush=True,
                        end="",
                        file=output_file,
                    )

    render_questions()
    item_sum: int = 0

    if not multi_line:
        for i in items:
            item_sum += len(i) + 3

        item_sum -= 3

    while True:
        key = getch(input_file)

        if key == CTRL_C:
            raise KeyboardInterrupt

        if key in {"A", "D"}:
            if selected != 0:
                selected -= 1
            else:
                continue
        elif key in {"B", "C"}:
            if selected != len(items) - 1:
                selected += 1
            else:
                continue
        elif key == "\r":
            if transient:
                print(
                    left(item_sum) + " " * item_sum + left(item_sum),
                    end="",
                    flush=True,
                    file=output_file,
                )
            if not multi_line:
                print(file=output_file)

            return selected
        else:
            continue

        if multi_line:
            for _ in items:
                print(
                    PREVIOUS_FIRST,
                    end="",
                    flush=True,
                    file=output_file,
                )
        else:
            print(left(item_sum), end="", flush=True, file=output_file)

        render_questions()


def ask_radio(
    question: str,
    items: Sequence[str],
    *,
    input_file: IO[str] | None = None,
    output_file: IO[str] | None = None,
    selected_color: str = Fore.GREEN + Style.DIM + UNDERLINE,
) -> int:
    """
    Display a radio selection with a prompt.

    Args:
        items: Sequence of available selections. Must have at least 2 values.
        input_file: File descriptor to read input from. `sys.stdin` by default.
        output_file: File descriptor to write output to. `sys.stdout` by default.
        selected_color: Color to use for selected values. Green by default.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C

    Returns:
        Index of the selected item.
    """
    with Prompt(question, output_file=output_file) as prompt:
        index = radio(
            items,
            input_file=input_file,
            output_file=output_file,
            selected_color=selected_color,
            transient=True,
        )
        prompt.answer = items[index]
        prompt.done()
        return index


def confirm(
    *,
    default: Literal[True, False] = True,
    input_file: IO[str] | None = None,
    output_file: IO[str] | None = None,
    selected_color: str = Fore.GREEN + Style.DIM + UNDERLINE,
) -> bool:
    """
    Get a yes or no value from the user, as a boolean.

    Args:
        default: Option to be the selected value at first.
        input_file: File descriptor to read input from. `sys.stdin` by default.
        output_file: File descriptor to write output to. `sys.stdout` by default.
        selected_color: Color to use for selected values. Green by default.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C

    Returns:
        `True` if yes was selected, `False` if no was selected.
    """
    seq = ["Yes", "No"] if default is True else ["No", "Yes"]
    index = radio(
        seq,
        input_file=input_file,
        output_file=output_file,
        selected_color=selected_color,
        transient=True,
    )
    return seq[index] == "Yes"


def ask_confirm(
    question: str,
    *,
    default: Literal[True, False] = True,
    input_file: IO[str] | None = None,
    output_file: IO[str] | None = None,
    selected_color: str = Fore.GREEN + Style.DIM + UNDERLINE,
) -> bool:
    """
    Get a yes or no value from the user, with a prompt.

    Args:
        question: Prompt to display.
        default: Option to be the selected value at first.
        input_file: File descriptor to read input from. `sys.stdin` by default.
        output_file: File descriptor to write output to. `sys.stdout` by default.
        selected_color: Color to use for selected values. Green by default.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C

    Returns:
        `True` if yes was selected, `False` if no was selected.
    """
    with Prompt(question, output_file=output_file) as prompt:
        value = confirm(
            default=default,
            input_file=input_file,
            output_file=output_file,
            selected_color=selected_color,
        )
        prompt.answer = "Yes" if value else "No"
        prompt.done()
        return value



from __future__ import annotations
import os
import sys
from typing import IO


def _handle_read(val: str) -> str:
    if val == "":
        raise EOFError

    return val


if os.name == "nt":
    import msvcrt

    def getch(file: IO[str]):
        if file is sys.stdin:
            return msvcrt.getwch()
        return _handle_read(file.read(1))

else:
    import termios
    import tty

    def getch(file: IO[str]):
        fd = file.fileno()
        try:
            old_settings = termios.tcgetattr(fd)
        except termios.error:
            # running in a file stream or something similar, just read
            return _handle_read(file.read(1))

        try:
            tty.setraw(file.fileno())
            ch = file.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

        return ch



from __future__ import annotations

import sys
from typing import IO, Any

from colorama import Fore, Style, init
from typing_extensions import Final, Self

from ._styles import BOLD, CLEAR, PREVIOUS_FIRST, UNDERLINE, right

__all__ = ("Prompt",)


init()


_ARROW: Final[str] = f"{Style.BRIGHT}{Fore.BLACK}›{Style.RESET_ALL}"
_CHECK: Final[str] = f"{Style.BRIGHT}{Fore.GREEN}✔{Style.RESET_ALL}"
_ERROR: Final[str] = f"{Style.BRIGHT}{Fore.RED}✗{Style.RESET_ALL}"


class Prompt:
    """Low level prompt API."""

    def __init__(
        self,
        prompt_str: str,
        *,
        output_file: IO[str] | None = None,
        prompt_char: str = "?",
    ) -> None:
        self.prompt_str = prompt_str
        self.output_file = output_file or sys.stdout
        self.prompt_char = prompt_char
        self.answer: str = ""
        self.answer_color = Fore.WHITE
        self._shown_error = False

    def write(self, *messages: Any, nl: bool = False) -> None:
        print(
            *messages,
            flush=True,
            end="" if not nl else "\n",
            file=self.output_file,
        )

    def _make_prompt_str(self) -> str:
        return (
            f"{self.prompt_char}{Style.RESET_ALL}"
            + f" {BOLD}{self.prompt_str} {_ARROW} {Style.RESET_ALL}"
            + f"{self.answer_color}{self.answer}{Style.RESET_ALL}"
        )

    def render(self) -> str:
        prompt_str = self._make_prompt_str()
        self.write("\r" + prompt_str)
        return prompt_str

    def error(
        self,
        error: str,
        *,
        color: str = Style.BRIGHT + Fore.RED,
    ) -> None:
        self._shown_error = True
        self.prompt_char = _ERROR
        self.answer_color = Fore.RED + Style.BRIGHT + UNDERLINE
        self.write(PREVIOUS_FIRST + CLEAR)
        self.render()
        self.write(f"\n\r{_ARROW} {color}{error}{Style.RESET_ALL}")
        self.write(PREVIOUS_FIRST + right(len(self.prompt_str) + 5))

    def done(
        self,
    ) -> None:
        self.prompt_char = _CHECK
        self.answer_color = Fore.GREEN + Style.DIM + UNDERLINE
        prompt_str = self._make_prompt_str()
        self.write(
            PREVIOUS_FIRST + prompt_str,
            nl=True,
        )

    def finish(self) -> None:
        if self._shown_error:
            self.write(PREVIOUS_FIRST + CLEAR)

    def __enter__(self) -> Self:
        self.render()
        return self

    def __exit__(self, *_) -> None:
        self.finish()



from __future__ import annotations
import sys
from io import StringIO
from typing import IO, Any, Sequence

from colorama import Fore, Style

from ._getch import getch
from ._styles import ARROW, BACKSPACE, CTRL_C, ITALIC, left

__all__ = ("take_input",)


def take_input(
    file: IO[str] | None = None,
    *,
    output: IO[str] | None = None,
    autocomplete: Sequence[str] | None = None,
    min_autocomplete_chars: int = 2,
    initial: str | None = None,
    initial_color: str = Fore.BLACK + Style.BRIGHT + ITALIC,
    color_sequence: str = Fore.WHITE,
) -> str:
    """
    Args:
        file: File to read from. `sys.stdin` by default.
        output: Output file to print to. `sys.stdout` by default.
        autocomplete: List of strings available for autocomplete.
        min_autocomplete_chars: Minimum amount of characters to be entered before autocomplete shows up.
        initial: Initial value to show before a value is inputted.
        initial_color: Color of the initial value. Gray by default.
        color_sequence: Color of the inputted characters. White by default.

    Returns:
        The inputted string.

    Raises:
        KeyboardInterrupt: User pressed CTRL+C
    """
    buffer = StringIO()
    slice: str | None = None
    file = file or sys.stdin
    output = output or sys.stdout
    first_iteration = True

    def write(*messages: Any, nl: bool = False) -> None:
        print(
            *messages,
            end="" if not nl else "\n",
            flush=True,
            file=output,
        )

    if initial:
        write(initial_color + initial + Style.RESET_ALL + left(len(initial)))

    while True:
        key = getch(file)
        if first_iteration and initial:
            write(" " * len(initial) + left(len(initial)))

        first_iteration = False
        bufsize = buffer.tell()
        if (bufsize + 1 > min_autocomplete_chars) and autocomplete:
            val = buffer.getvalue() + key
            found = False

            for item in autocomplete:
                if item.startswith(val):
                    slice = item[bufsize:]
                    write(
                        Style.RESET_ALL
                        + ITALIC
                        + Style.BRIGHT
                        + Fore.BLACK
                        + slice
                        + Style.RESET_ALL
                    )

                    for _ in slice:
                        write("\b")

                    found = True
                    break

            if (not found) and slice:
                write(" " * len(slice))
                write(left(len(slice)))

        if key == CTRL_C:
            raise KeyboardInterrupt

        if key == ARROW:
            continue

        if (key == "\t") and slice:
            write(slice[1:])
            buffer.write(slice[1:])
            continue

        if key == BACKSPACE:
            index = bufsize - 1
            if index < 0:
                continue

            buffer.truncate(index)
            buffer.seek(index)
            if slice:
                write(" " * len(slice))
                write("\b" * len(slice))

            write("\b \b")
            continue

        if key in {"\r", "\n"}:
            write("\n")
            break

        write(color_sequence + key + Style.RESET_ALL)
        buffer.write(key)

    return buffer.getvalue()
